/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Demo how to use Algorithm 10.4 (Automatic-Rejection-Inversion), p. 226
 *
 * this file: demo_algo10_04.c
 *
 * to compile this file you can use the command:
 *
 * cc -o algo10_04 demo_algo10_04.c algo10_04.c statutil.c uniform.c -lm
 *
 */

#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include "uniform.h"
#include "algo.h"
#include "statutil.h"

#define WID 300000 /*samplesize that should be drawn for every distribution*/


#define SM_LINT -0x80000000 /*smallest long int on most 32-bit machines*/
#define LA_LINT  0x7FFFFFFF /*largest long int on most 32-bit machines*/

/*structure containing borders and mode of a distribution*/
typedef struct{ long int m/*mode*/,
                 r,l;/*smallest and largest integer with positive porbability*/
      	      }DOMEN;

/**************************************************************************/
#define P (*param)
/*computes a constant for Poisson histogramm*/
void histpsu(param)
PARAM *param;/*mu= param.d1*/ 
{ P.d2=log(P.d1); }
/*Poisson histogramm: */
/*before using histpsu must be called to compute some constants*/
double histp(param,k)
PARAM *param;/*mu= param.d1*/ 
long int k;
{  return exp(-P.d1+k*P.d2-lnkfakt(k));
}
/*************************************************************************/
/*computes constants for Binomial histogramm*/
void histbinsu(param)
PARAM *param;/*param.i1 n,  param.d1 p*/
{ P.d2=lnkfakt(P.i1);
  P.d3=log(P.d1);
  P.d4=log(1-P.d1);
}
/*histogram of binomial distribution: */
/*before using histbinsu must be called to compute some constants*/
double histbin(param,k)
PARAM *param;/*param.i1 n,  param.d1 p*/
long int k;
{ 
  return exp(P.d2-lnkfakt(k)-lnkfakt(P.i1-k)+
	     k*P.d3+(P.i1-k)*P.d4);
}
/*************************************************************************/
/*computes a constant for Hypergeometric histogramm*/
void histhypsu(param)
PARAM *param;/*param.i1 N,  param.i2 M, param.i3 n*/
{ P.d1=lnkfakt(P.i3)-lnkfakt(P.i1)+lnkfakt(P.i2)+
       lnkfakt(P.i1-P.i2)+lnkfakt(P.i1-P.i3);}
/*Hypergeometr. histogramm:*/
/*before using histhypsu must be called to compute some constants*/
double histhyp(param,k)
PARAM *param;/*param.i1 N,  param.i2 M, param.i3 n*/
long int k;
{  return exp(P.d1-lnkfakt(k)-lnkfakt(P.i2-k)-lnkfakt(P.i3-k)-
	      lnkfakt(P.i1-P.i2-P.i3+k));
}
/*************************************************************************/
/*computes constants for negative Binomial histogramm*/
void histnbsu(param)
PARAM *param;/*param.i1 n,  param.d1 p*/
{ 
  P.d2=P.i1*log(P.d1)-lnkfakt(P.i1-1);
  P.d3=log(1-P.d1);
}
/*Negativ Binomial histogramm:*/
/*before using histnbsu must be called to compute some constants*/
double histnb(param,k)
PARAM *param;/*param.i1 n,  param.d1 p*/
long int k;
{ 
  return exp(lnkfakt(P.i1+k-1)-lnkfakt(k)+P.d2+k*P.d3);
}
/**************************************************************************/
/*histogramm of Zipf-3-distribution, P(k)=const/k^3*/
double histzipf3(param,k)
PARAM *param;
long int k;
{ static double konst=0.831907372581 /* = 1/zeta[3] */;
  return konst*exp(-log((double)k)*3.);
}
/**************************************************************************/
/*proportional to histogramm of Zipf-2-distribution P(k)=const/k^2*/
double histzipf2prop(param,k)
PARAM *param;
long int k;
{ 
  return exp(-log((double)k)*2.);
}
/*************************************************************************/
/*computes for the above histograms the mode and domain of the distribution*/
void compdom(h,domain,param)
DOMEN *domain;
double (*h)();
PARAM *param;/*die Parameter der Verteilung, Aufruf von h als (*h)(param,k)*/
{ long int help;
  (*domain).l=SM_LINT;
  (*domain).r=LA_LINT;
  if (h==histp)     
    { (*domain).m=(long int)(P.d1);
      (*domain).l=0;
    }
  if (h==histbin)
    { (*domain).m=(long int)((P.i1+1)*P.d1);
      (*domain).r=(P.i1);
      (*domain).l=0;
    }
  if (h==histhyp)
    { (*domain).m=(long int)((P.i3+1.)*(P.i2+1.)/(P.i1+2.));
      if (P.i2<=P.i3) (*domain).r=(P.i2);
      else (*domain).r=(P.i3);
      help=P.i3-P.i1+P.i2;
      if (help<=0) (*domain).l=0;
      else (*domain).l=help;
    }
  if (h==histnb)
    { if (P.i1*(1.-P.d1)<=1) (*domain).m=0;
      else (*domain).m=ceil((P.i1*(1.-P.d1)-1.)/P.d1);
      (*domain).l=0;
    }
  if (h==histzipf3||h==histzipf2prop)
    { (*domain).m=1;
      (*domain).l=1;
    }
}
/*************************************************************************/
/*computes the chi-square statistic*/
double chi2discr(classes,obs,wid,domain,h,param,output)
long int classes/*number of classes*/,obs[]/*vector of counted values*/,
     wid/*number of repetitions*/;
DOMEN *domain;
double (*h)();
PARAM *param;/*paramters of the distr., call of the prob(k) as (*h)(param,k)*/
char output/*(0.. no, 1... result, 2.. more)*/;
{ double chi2=0.,ex,sumex=0.,sumob=0.,ob;
  long int i,b,a,cl;
 cl=classes+classes;
 do{ cl=cl/2;
     b=(*domain).m-cl/2;
     if(b<(*domain).l) b=(*domain).l;
     a=b+cl-1;
     if(a>(*domain).r) a=(*domain).r;
   }while((*h)(param,a)*wid<5||(*h)(param,b)*wid<5);


for(i=b;i<a;i++)
  { ex=(*h)(param,i)*wid;
    ob=obs[i];    
    sumex+=ex;
    sumob+=ob;
    chi2+= (ob-ex)*(ob-ex)/ex;
    if(output>=2) printf("for %ld: exp %.2f ob %.0f\n",i,ex,ob); 
  }
  ob=wid-sumob;
  ex=wid-sumex; 
  chi2+= (ob-ex)*(ob-ex)/ex;
  if(output>=2) printf("All other numbers: exp %.2f ob %.0f\n",ex,ob);
  if(output>=1) 
    printf("for samplesize n=%ld  and %ld DF: P-Value  %f\n",wid,a-b,cdf_chia(chi2,a-b));
  return(chi2);
}
/*************************************************************************/



#define ARRSIZE 10000

long kcount[ARRSIZE];

int main()
{ long int i,k;
  PARAM par;
  DOMEN domain;
  TOTAL *poi;

 setseed(132213248); /*set seed for uniform generator*/

  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nPoisson distribution with mu=50.\n");
  par.d1=50.;
  histpsu(&par);
  compdom(histp,&domain,&par);
  poi=setup10_04(histp,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=generator10_04(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;}
  free10_04(poi);
  chi2discr(50,kcount,WID,&domain,histp,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nBinomial distribution with p=0.3 and n=200\n");
  par.d1=0.3;
  par.i1=200;
  histbinsu(&par);
  compdom(histbin,&domain,&par);
  poi=setup10_04(histbin,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=generator10_04(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free10_04(poi);
  chi2discr(50,kcount,WID,&domain,histbin,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nHypergeometric distribution with N=1000 M=100 n=90\n");
  par.i1=1000;
  par.i2=100;
  par.i3=90;
  histhypsu(&par);
  compdom(histhyp,&domain,&par);
  poi=setup10_04(histhyp,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=generator10_04(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free10_04(poi);
  chi2discr(50,kcount,WID,&domain,histhyp,&par,1);
  for(i=0;i<ARRSIZE;i++) kcount[i]=0;
printf("\nNegative binomial distribution with n=3 p=0.7\n");
  par.i1=3;
  par.d1=0.7;
  histnbsu(&par);
  compdom(histnb,&domain,&par);
  poi=setup10_04(histnb,&par,domain.m,domain.l,domain.r,-1.);
  for(i=0;i<WID;i++) 
    { k=generator10_04(poi);
      if (k>=0&&k<ARRSIZE) kcount[k]++;
    }
  free10_04(poi);
  chi2discr(50,kcount,WID,&domain,histnb,&par,1);


  return 0;
}    
 
